/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2010 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.persistence.ContainerDataIterator;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.MultiContainerDataIterator;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.IndexResult;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.IOException;
import java.util.*;

/**
 * Task zur Bearbeitung von Archiv-Informationsanfragen.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 8678 $ / $Date: 2011-01-27 15:38:11 +0100 (Do, 27 Jan 2011) $ / ($Author: rs $)
 */
public class ArchiveInfoQueryTask extends QueryTask {

	private static final boolean GAP = true, NO_GAP = false, NO_DIR_ACCESS = false;

	private static final int NO_MEDIUM_ID = -1;

	private ArchiveDataSpecification[] adsList;

	private byte[] indexMem;

	private int entryCounter;

	private DummyTask dummyTask = new DummyTask();	// zum autom., Verlaengern des Loeschschutzes

	public ArchiveInfoQueryTask(ArchiveManager archiveMgr, MultiTaskManager tMgr) {
		super(archiveMgr, tMgr);
		logger = Debug.getLogger();
	}

	private void analyze() throws Exception {
		createQueryData();
		adsList = parseArchiveDataSpec(deserializer);
	}

	private IndexResult getIndexResult(long obj, long atg, long asp, short sv, ArchiveTimeSpecification ats, DataIdentNode din, ArchiveDataKind adk)
			throws Exception {
		if(!din.indexesAssigned(adk)) persMgr.assignIndexes(didTree, obj, atg, asp, sv, adk, this, indexMem);
		if(ats.getTimingType().equals(TimingType.DATA_INDEX)) {
			return din.getDataIdxContIDs(adk, persMgr.getPathName(obj, atg, asp, sv, adk), indexMem, ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else if(ats.getTimingType().equals(TimingType.DATA_TIME)) {
			return din.getDataTimeContIDs(adk, persMgr.getPathName(obj, atg, asp, sv, adk), indexMem, ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else if(ats.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
			return din.getArSTimeContIDs(adk, persMgr.getPathName(obj, atg, asp, sv, adk), indexMem, ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else {
			return IndexResult.EMPTY;
		}
	}

	/**
	 * Errechnet den groessten Datenindex aus allen Containerlisten, der unter dem angegebenen TimingType noch innerhalb des angefragten Bereichs liegt.
	 *
	 * @param obj ObjektID
	 * @param atg AttributgruppenID
	 * @param asp AspektID
	 * @param sv  imVar
	 * @param ats Bereich
	 * @param oa  Liste der OA-Container (mit Hilfe der Indexe ermittelt)
	 * @param on  Liste der ON-Container (mit Hilfe der Indexe ermittelt)
	 * @param na  Liste der NA-Container (mit Hilfe der Indexe ermittelt)
	 * @param nn  Liste der NN-Container (mit Hilfe der Indexe ermittelt)
	 *
	 * @return groessten passenden Datenindex
	 *
	 * @throws PersistenceException
	 */
	private long calcLastDIdx(
			long obj, long atg, long asp, short sv, ArchiveTimeSpecification ats, IndexResult oa, IndexResult on, IndexResult na, IndexResult nn)
			throws PersistenceException {

		if(ats.getTimingType().equals(TimingType.DATA_INDEX)) return ats.getIntervalEnd();

		IndexResult[] ir = new IndexResult[]{oa, on, na, nn};
		long maxDidx = -1;
		for(int archiveDataKindIndex = 0; archiveDataKindIndex < 4; archiveDataKindIndex++) {
			if(ir[archiveDataKindIndex] != null && !ir[archiveDataKindIndex].isEmpty()) {
				final String pathName = persMgr.getPathName(obj, atg, asp, sv, archiveDataKindIndex);

				if(ats.getTimingType().equals(TimingType.DATA_TIME)) {
					// Die Container im IndexResult sind nach DTMin sortiert, deswegen kann
					// der letzte DS in allen Containern liegen
					for(int j = 0; j < ir[archiveDataKindIndex].size(); j++) {
						final String fileName = ContainerFile.getContainerFileName(ir[archiveDataKindIndex].getCID(j));
						maxDidx = Math.max(maxDidx, getMaxContainerDataIndexMaxDataTime(ats.getIntervalEnd(), new File(pathName, fileName)));
					}
				}
				else if(ats.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
					// Die Archivzeit ist monoton ueber die Container, also muss der letzte DS
					// im letzten Container liegen:
					final String fileName = ContainerFile.getContainerFileName(ir[archiveDataKindIndex].getCID(ir[archiveDataKindIndex].size() - 1));
					maxDidx = Math.max(maxDidx, getMaxContainerDataIndexMaxArchiveTime(ats.getIntervalEnd(), new File(pathName, fileName)));
				}
			}
		}
		return maxDidx;
	}

	/** Gibt von einer Containerdatei den hchsten DataIndex zurck, von dem die DataTime nicht lter als die angegebene Zeit ist
	 *
	 * @param maxTime Maximalzeit der Datenindexe, die bercksichtigt werden
	 * @param containerFile Die zu ffnende Containerdatei
	 * @return Den maximalen Datenindex innerhalb der Containerdatei oder -1 falls die Containerdatei gelscht ist oder keine Daten enthlt
	 * @throws PersistenceException Fehler in der Persistenz
	 */
	private long getMaxContainerDataIndexMaxDataTime(final long maxTime, final File containerFile) throws PersistenceException {
		long maxDataIndex = -1;
		persMgr.accessContainerReadOnly(this, containerFile);
		try {
			persMgr.readContainerHeader(this);
			final boolean markedAsDeleted = persMgr.getContainerHeaderParamAsBoolean(this, ContainerHdr.CHP_DELETED);

			final long maxDataIndexFromHeader = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_DATA_IDX_MAX);
			final long maxContainerTime = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_DATA_TIME_MAX);

			//Versuchen den maximalen Datensatz ber Header zu ermitteln
			if((maxDataIndexFromHeader > 0 && (maxContainerTime < maxTime)) || markedAsDeleted) return maxDataIndexFromHeader;

			// Falls der Container nicht abgeschlossen ist, oder die Maximalzeit kleiner als die maximale Datensatzzeit im Container ist,
			// mssen die Datenstze einzeln durchgegangen werden
			ContainerDataIterator it = persMgr.getArchiveData(this);

			while(it.readNextDataHdr()) {
				if(it.getDataTime() <= maxTime) maxDataIndex = Math.max(maxDataIndex, it.getDataIndex());
				it.skipData();
			}
		}
		finally {
			persMgr.leaveContainer(this);
		}
		return maxDataIndex;
	}

	/** Gibt von einer Containerdatei den hchsten DataIndex zurck, von dem die ArchiveTime nicht lter als die angegebene Zeit ist
	 *
	 * @param maxTime Maximalzeit der Datenindexe, die bercksichtigt werden
	 * @param containerFile Die zu ffnende Containerdatei
	 * @return Den maximalen Datenindex innerhalb der Containerdatei oder -1 falls die Containerdatei gelscht ist oder keine Daten enthlt
	 * @throws PersistenceException Fehler in der Persistenz
	 */
	private long getMaxContainerDataIndexMaxArchiveTime(final long maxTime, final File containerFile) throws PersistenceException {
		long maxDataIndex = -1;
		persMgr.accessContainerReadOnly(this, containerFile);
		try {
			persMgr.readContainerHeader(this);
			final boolean markedAsDeleted = persMgr.getContainerHeaderParamAsBoolean(this, ContainerHdr.CHP_DELETED);

			//Versuchen den maximalen Datensatz ber Header zu ermitteln
			final long maxDataIndexFromHeader = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_DATA_IDX_MAX);
			final long maxContainerTime = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_ARC_TIME_MAX);

			if((maxDataIndexFromHeader > 0 && (maxContainerTime < maxTime)) || markedAsDeleted) return maxDataIndexFromHeader;

			// Falls der Container nicht abgeschlossen ist, oder die Maximalzeit kleiner als die maximale Datensatzzeit im Container ist,
			// mssen die Datenstze einzeln durchgegangen werden
			ContainerDataIterator it = persMgr.getArchiveData(this);

			while(it.readNextDataHdr()) {
				if(it.getArcTime() <= maxTime) maxDataIndex = Math.max(maxDataIndex, it.getDataIndex());
				it.skipData();
			}
		}
		finally {
			persMgr.leaveContainer(this);
		}
		return maxDataIndex;
	}


	private boolean isGap(long start, ContainerDataIterator it) {
		if(!it.isHdrMaxVal()) {
			return start == -1 ? false : Util.dIdxNoModBits(it.getDataIndex()) - Util.dIdxNoModBits(start) > 1;
		}
		else {
			return false;
		}
	}


	/**
	 * Beantwortet die Archivinformationsanfragen. Die einzelnen Archivinformationsanfragen aus {@link #adsList} werden bearbeitet und beantwortet.
	 *
	 * @throws IOException
	 * @throws DataNotSubscribedException
	 * @throws ConfigurationException
	 * @throws SendSubscriptionNotConfirmed
	 */
	private void sendResponse() throws IOException, DataNotSubscribedException, ConfigurationException, SendSubscriptionNotConfirmed {
		indexMem = new byte[INDEX_MEMORY_SIZE];
		bosResult.reset();	        // genuegt, man muss nicht jedesmal einen neuen Serialisierer anlegen

		serializer.writeByte(1);		// erfolgreich (nehmen wir erstmal an...)
		serializer.writeInt(0);			// Groesse Liste (wird spter besetzt)
		entryCounter = 0;

		DataIdentNode din = null;
		IndexResult[] idxRes = new IndexResult[4];
		int adsCounter = 0;
		boolean criticalAreaEntered = false;
		MultiContainerDataIterator mci = null;

		for(ArchiveDataSpecification ads : adsList) {
			suspendTaskIfNecessary();		//Task anhalten, falls Archiv berlastet.
			if(isTaskTerminated()) break;

			long obj = ads.getObject().getId();
			long atg = ads.getDataDescription().getAttributeGroup().getId();
			long asp = ads.getDataDescription().getAspect().getId();
			short sv = (short)ads.getDataDescription().getSimulationVariant();
			ArchiveTimeSpecification ats = ads.getTimeSpec();
			HashSet<Long> ls_extended = new HashSet<Long>();

			try {
				if(ats.isStartRelative()) throw new Exception("Relative Intervallgrenzen sind in Archiv-Informationsanfragen nicht erlaubt.");

				// Synchronisierung ueber DataIdentNode:
				din = didTree.existsCreate(ads); // not null
				if(din.enterCriticalArea()) {
					criticalAreaEntered = true;

					if(ads.getDataKinds().isOnline()) {
						if(!din.indexesAssigned(OA)) persMgr.assignIndexes(didTree, ads, OA, this, tempMem);
						idxRes[0] = getIndexResult(obj, atg, asp, sv, ats, din, OA);
					}
					if(ads.getDataKinds().isOnlineDelayed()) {
						if(!din.indexesAssigned(ON)) persMgr.assignIndexes(didTree, ads, ON, this, tempMem);
						idxRes[1] = getIndexResult(obj, atg, asp, sv, ats, din, ON);
					}
					if(ads.getDataKinds().isRequested()) {
						if(!din.indexesAssigned(NA)) persMgr.assignIndexes(didTree, ads, NA, this, tempMem);
						idxRes[2] = getIndexResult(obj, atg, asp, sv, ats, din, NA);
					}
					if(ads.getDataKinds().isRequestedDelayed()) {
						if(!din.indexesAssigned(NN)) persMgr.assignIndexes(didTree, ads, NN, this, tempMem);
						idxRes[3] = getIndexResult(obj, atg, asp, sv, ats, din, NN);
					}
					din.exitCriticalArea();
					criticalAreaEntered = false;

					// der allerletzte, von dieser Anfrage betroffene Datenindex
					long finalDIdx = calcLastDIdx(obj, atg, asp, sv, ats, idxRes[0], idxRes[1], idxRes[2], idxRes[3]);

					mci = new MultiContainerDataIterator(
							archMgr, obj, atg, asp, sv, idxRes[0], idxRes[1], idxRes[2], idxRes[3], MultiContainerDataIterator.DEL_MINMAX, this
					);
					ContainerDataIterator contIt;
					TimingType tt = ats.getTimingType();

					long curDIdx = -1, curATime = -1, curDTime = -1;						// Werte des aktuellen Durchlaufs
					long lastDIdx = -1, lastDTime = -1, lastATime = -1, lastContID = -1;	// Werte des vorherigen Durchlaufs
					long lbDIdx = -1, lbATime = -1, lbDTime = -1;							// Untere Grenze der Ausgabezeile
					int lastMedienID = -1, curMedienID = -1;							// letzte und aktuelle (wg. Plattenzugriff) medienID
					boolean lastDirAccess = true, curDirAccess = true;				// letzter und aktueller (wg. Plattenzugriff) Direktzugriff

					boolean blockAlreadyWritten;

					// Hauptschleife:
					while(true) {
						if(isTaskTerminated()) throw new Exception("Das Archivsystem wurde beendet.");

						contIt = mci.readSkipNextDIdx();
						if(contIt == null) break;

						curDIdx = contIt.getDataIndex();
						curATime = contIt.getArcTime();
						curDTime = contIt.getDataTime();

						if(finalDIdx != -1 && curDIdx > finalDIdx) break;

						if(lbDIdx == -1) {
							// Untergrenzen initialisieren
							lbDIdx = curDIdx;
							lbATime = curATime;
							lbDTime = curDTime;
						}
						blockAlreadyWritten = false;
						if(mci.getCurContID() != lastContID) {
							// erster oder neuer Container
							lastMedienID = curMedienID;
							curMedienID = getMedienID(mci);
							lastDirAccess = curDirAccess;
							curDirAccess = getDirAccess(mci);

							if(lastContID != -1 && (lastMedienID != curMedienID || lastDirAccess != curDirAccess)) {
								// neuer Container (mindestens der 2.) mit anderen access-/medienID-Werten
								addEntry(lbDIdx, lbATime, lbDTime, lastDIdx, lastATime, lastDTime, tt, NO_GAP, lastDirAccess, lastMedienID, adsCounter);

								lbDIdx = curDIdx;
								lbATime = curATime;
								lbDTime = curDTime;
								blockAlreadyWritten = true;
							}
							lastContID = mci.getCurContID();
						}

						if(isGap(lastDIdx, contIt)) {
							if(!blockAlreadyWritten )
							// Zeile fuer Zustand bis zur Luecke (falls nicht schon wg. Containerwechsel ausgegeben):
							{
								addEntry(lbDIdx, lbATime, lbDTime, lastDIdx, lastATime, lastDTime, tt, NO_GAP, curDirAccess, curMedienID, adsCounter);
							}

							// Zeile fuer die Luecke:
							addEntry(lastDIdx, lastATime, lastDTime, curDIdx, curATime, curDTime, tt, GAP, NO_DIR_ACCESS, NO_MEDIUM_ID, adsCounter);

							// naechste Zeile geht beim Lueckenende los:
							lbDIdx = curDIdx;
							lbATime = curATime;
							lbDTime = curDTime;
						}
						lastDIdx = curDIdx;
						lastATime = curATime;
						lastDTime = curDTime;

						// Archivinformationsanfrage loest automatische Loeschschutzverlaengerung aus (gemaess TAnfArS)
						if(!ls_extended.contains(mci.getCurContID())) {
							extendLoeschschutz(obj, atg, asp, sv, contIt.getArchiveDataKind(), mci.getCurContID());
							ls_extended.add(mci.getCurContID());
						}
					}

					// letzten Eintrag schreiben:
					addEntry(lbDIdx, lbATime, lbDTime, curDIdx, curATime, curDTime, tt, NO_GAP, curDirAccess, curMedienID, adsCounter);
				}
				else {
					throw new Exception("Das Archivsystem wurde beendet.");
				}
			}
			catch(Exception e) {
				logger.error("Fehler beim Bearbeiten der Archivinformationsanfrage", e);
				printError("Fehler beim Bearbeiten der Archivinformationsanfrage: " + e.getMessage());
				sendResultData(bosResult.toByteArray(), MessageType.QUERY_INFO_RESULT);
				return;
			}
			finally {
				if(din != null && criticalAreaEntered) din.exitCriticalArea();
				closeMultiContIter(mci);
			}
			adsCounter++;
		}
		sendResultData(insertNumOfEntries(entryCounter), MessageType.QUERY_INFO_RESULT);
	}

	private void extendLoeschschutz(long obj, long atg, long asp, int sv, ArchiveDataKind adk, long contID) throws PersistenceException {
		try {
			persMgr.accessContainer(dummyTask, obj, atg, asp, sv, adk, contID);
			if(persMgr.isContainerClosed(dummyTask)) {
				persMgr.setAutoExtendLoeschutz(dummyTask, RestoreTask.getLOESSCHUTZ_VERL(), t_lsMax);
				persMgr.writeContainerHeader(dummyTask);
			}
		}
		finally {
			persMgr.leaveContainer(dummyTask);
		}
	}

	private void closeMultiContIter(MultiContainerDataIterator multiContIt) {
		if(multiContIt != null) multiContIt.close();
	}

	private boolean getDirAccess(MultiContainerDataIterator mci) throws PersistenceException {
		return !persMgr.getContainerHeaderParamAsBoolean(this, mci.getCurContFileIdx(), ContainerHdr.CHP_DELETED);
	}

	private int getMedienID(MultiContainerDataIterator mci) throws PersistenceException {
		String hdrMedID = persMgr.getContainerHeaderParamAsString(this, mci.getCurContFileIdx(), ContainerHdr.CHP_MEDIUM_ID);
		if(!hdrMedID.equals(ContainerHdr.MEDIUM_UNSAVED) && !hdrMedID.equals(ContainerHdr.MEDIUM_DESTROYED)) {
			return Integer.parseInt(hdrMedID);
		}
		else {
			return -1;
		}
	}

	private long ttVal(TimingType tt, long dIdx, long aTime, long dTime) {
		return tt.equals(TimingType.DATA_TIME) ? dTime : (tt.equals(TimingType.ARCHIVE_TIME) ? aTime : dIdx);
	}

	private byte[] insertNumOfEntries(int numOfEntries) throws IOException {
		byte[] resultBytes = bosResult.toByteArray();
		bosResult.reset();
		serializer.writeInt(numOfEntries);
		System.arraycopy(bosResult.toByteArray(), 0, resultBytes, 1, 4);
		return resultBytes;
	}

	private void printError(String msg) throws IOException {
		bosResult.reset();
		serializer.writeByte(0);	// Fehler
		serializer.writeString(msg);
	}

	private void addEntry(
			long startDIdx, long startAT, long startDT, long endDIdx, long endAT, long endDT, TimingType tt, boolean gap, boolean direct, int volID, int idx)
			throws IOException {
		// System.out.println(startDIdx + "-" + endDIdx + " " + (gap ? "GAP":"DAT")+ " " + (direct ? "dac":"nodac"));
		serializer.writeLong(ttVal(tt, startDIdx, startAT, startDT));	  // 3.
		serializer.writeLong(ttVal(tt, endDIdx, endAT, endDT));	  // 4.
		serializer.writeByte(tt.equals(TimingType.DATA_TIME) ? 1 : (tt.equals(TimingType.ARCHIVE_TIME) ? 2 : 3));	// 5. TT
		serializer.writeByte(gap ? 1 : 0);		// 6. Gap
		serializer.writeByte(direct ? 1 : 0);	// 7. Access
		serializer.writeInt(volID);				// 8. volID
		serializer.writeInt(idx);			    // 9.
		entryCounter++;
	}

	public void work() {
		//logger.finer(getName() + " Erhaelt Task: ", resultData);
		init();
		try {
			analyze();
			subscribeSender();
			// System.out.println("ArchiveInfoQueryTask.work");

			final SendingState state;
			// Nach 120 Sekunden ohne Sendesteuerung wird der Auftrag abgebrochen
			final long timeoutTime = System.currentTimeMillis() + (120 * 1000);
			synchronized(syncObj) {
				while(sendingState == SendingState.STOP) {
					final long delta = timeoutTime - System.currentTimeMillis();
					if(delta <= 0) break;
					syncObj.wait(delta);
				}
				state = sendingState;
			}

			if(state == SendingState.SEND) {
				sendResponse();
			}
		}
		catch(InterruptedException e) {
			logger.finer(getName() + " Task durch Interrupt beendet.");
		}
		catch(Exception e) {
			// resultData wird vor Aufruf der Debug-Methode in einen String konvertiert, um verschachtelte Debug-Aufrufe mit Deadlock-Gefahr zu vermeiden
			logger.error(getName() + " Fehler beim Bearbeiten der Archiv-Informationsanfrage: " + e.getMessage() + ", Anfrage: " + resultData.toString());
		}
		finally {
			unsubscribeSender(); // Als Sender der Archiv-Antwort abmelden
			persMgr.removeContainerFileObjects(this, MIN_CONTAINER_FILES); // ContainerFiles aufraeumen
			adsList = null;
			indexMem = null;
			bosResult.reset();
		}
	}


	private static class DummyTask extends Task {

		public DummyTask() {
			super(null, null);
		}
	}
}
