/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.backup.BackupModul;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.tasks.tasklistener.RingOfDeathListener;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.KeyValParam;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.util.*;

/**
 * Task zum Abgleich von Verwaltungsinformationen in den Container-Headern.
 * <p/>
 * Dieses Modul dient dem Abgleich von Verwaltungsinformation (=Container-Header) zwischen den Containern eines Mediums des Typs B und den entsprechenden
 * Containern im Persistenz-Verzeichnis des Archivsystems.<br> <br>
 * <p/>
 * Bei jedem Aufruf werden die Header aller Container des Archivsystems durch die entsprechenden Container-Header, die auf dem spezifizierten Medium vorhanden
 * sind, ersetzt.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision: 6419 $ / $Date: 2009-03-10 23:15:57 +0100 (Di, 10 Mrz 2009) $ / ($Author: rs $)
 */
public class HeaderRestoreTask extends SingleClientSenderTask {

	private final static String TMPDIR = System.getProperty("java.io.tmpdir");

	public HeaderRestoreTask(ArchiveManager archiveMgr, RingOfDeathListener rodl, RingBuffer<ResultData> inputData) throws ConfigurationException {
		super(archiveMgr, rodl, inputData);
		logger = Debug.getLogger();
	}

	protected void work(ResultData resultData) {
		if(!resultData.hasData() || isTaskTerminated()) return;
		try {
			logger.info("Starte Abgleich der Verwaltungsinformationen");
			init(resultData.getData());
			restoreHeader(resultData.getData());
			logger.info("Abgleich der Verwaltungsinformationen beendet");
			sendSuccessResponse();
		}
		catch(MediumNotAccessibleException e) {
			logger.warning("Abgleich der Verwaltungsinformationen fehlgeschlagen: Medium nicht verfgbar." + e.getMessage());
			MessageSender.getInstance().sendMessage(
					"", MessageType.APPLICATION_DOMAIN, "", MessageGrade.WARNING, MessageState.MESSAGE, "Bitte legen Sie Medium ein"
			);
			sendErrorResponse("Abgleich der Verwaltungsinformationen fehlgeschlagen: Medium nicht verfgbar." + e.getMessage());
		}
		catch(Exception e) {
			logger.warning("Abgleich der Verwaltungsinformationen fehlgeschlagen." + e.getMessage());
			sendErrorResponse("Abgleich der Verwaltungsinformationen fehlgeschlagen: " + e.getMessage());
		}
	}


	void restoreHeader(Data data) throws Exception {
		BackupModul backup = BackupTask.getBackupModul();

		int mediumID = deserializer.readInt();

		List<String> contFileNames = backup.getAllContFileNames(this, mediumID, -1);

		for(String contFile : contFileNames) {

			suspendTaskIfNecessary();	//Task anhalten, falls Archivsystem in kritischem Bereich			
			if(isTaskTerminated()) throw new Exception("Archivsystem wurde beendet. Verwaltungsinformationsabgleich abgebrochen");

			logger.finer("Checking file " + contFile + " ...");
			File tmpContFile = new File(TMPDIR, contFile);
			InputStream is = backup.restoreContainer(mediumID, contFile);
			FileOutputStream fos = new FileOutputStream(tmpContFile);
			Util.copyStreams(is, fos);
			is.close();
			fos.close();
			boolean busyAreaEntered = false;
			DataIdentNode din = null;
			try {
				// temporres File auslesen, um DID + ContainerID zu bekommen: 
				persMgr.accessContainerReadOnly(this, tmpContFile);
				persMgr.readContainerHeader(this);

				long objId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_OBJ_ID);
				long atgId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_ATG_ID);
				long aspId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_ASP_ID);
				int sv = persMgr.getContainerHeaderParamAsInt(this, ContainerHdr.CHP_SIM_VAR);
				ArchiveDataKind datakind = persMgr.getContainerHeaderParamAsArchiveDataKind(this, ContainerHdr.CHP_DATA_KIND);
				long contId = persMgr.getContainerHeaderParamAsLong(this, ContainerHdr.CHP_CONT_ID);
				persMgr.leaveContainer(this);

				// Pruefen, ob Zugriff auf Typ-A-Container erlaubt ist. Synchronisierung ueber DataIdentNode:
				din = archMgr.getDataIdentTree().existsCreate(objId, atgId, aspId, sv);
				busyAreaEntered = din.enterBusyArea();
				if(!busyAreaEntered) {
					logger.warning("Abgleich der Verwaltungsinformationen fehlgeschlagen fuer Datei " + contFile);
					continue; // Dieses File ueberspringen da Zugriff nicht erlaubt wurde.
				}

				// auf Container zugreifen
				persMgr.accessContainer(this, objId, atgId, aspId, sv, datakind, contId);

				/**
				 * Aus der Spec: Beim Abgleich kann es drei Situationen geben, die wie folgt behandelt werden: <br>
				 * <ul>
				 * <li>Der Container ist im Archivsystem nicht vorhanden: In diesem Fall wird der Container-Header in
				 * das Archivsystem kopiert und dort als gelschter und gesicherter Container markiert. Falls notwendig,
				 * wird ein entsprechendes Verzeichnis angelegt.</li>
				 * <li>der Container ist im Archivsystem vorhanden, allerdings als gelscht markiert (d.h. es ist nur
				 * der Container-Header vorhanden). In diesem Fall wird dieser Container-Header durch den
				 * Container-Header vom Typ-B-Medium ersetzt und als gelschter und gesicherter Container markiert. (Ein
				 * Test auf Gleichheit ist unntig).</li>
				 * <li>Der Container ist im Archivsystem ungelscht vorhanden. In diesem Fall werden die Header
				 * verglichen: Sind die Header vollstndig identisch (modulo Reihenfolge der Schlsselwrter), passiert
				 * nichts weiter. Sind sie nicht identisch, wird der Container gelscht und der Container-Header durch
				 * den Container-Header vom Typ-B-Medium ersetzt und als gelschter und gesicherter Container markiert.</li>
				 * </ul>
				 */

				// Fallunterscheidung:
				if(!persMgr.existsContainer(this)) {
					// 1. Container ist im Archiv nicht vorhanden
					// --> kopieren und als geloescht+gesichert markieren
					persMgr.createContainer(this, true);
					// persMgr.readContainerHeader(this); AS: nicht notwendig, createContainer() setzt headerRead=true
					persMgr.restoreHeader(this, tmpContFile);
					persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_DELETED, "1");
					persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_MEDIUM_ID, mediumID);
					persMgr.writeContainerHeader(this);
				}
				else {
					persMgr.readContainerHeader(this);
					boolean isDeleted = persMgr.getContainerHeaderParamAsBoolean(this, ContainerHdr.CHP_DELETED);
					if(isDeleted) {
						// 2. Container ist vorhanden, aber als geloescht markiert
						// --> Header ersetzen und als geloescht+gesichert markieren
						persMgr.restoreHeader(this, tmpContFile);
						persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_DELETED, "1");
						persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_MEDIUM_ID, mediumID);
						persMgr.writeContainerHeader(this);
					}
					else {
						// 3. Container ist ungeloescht vorhanden
						// --> Header vergleichen
						if(!isHeaderIdentical(tmpContFile)) {
							// --> Container loeschen, Header ersetzen und als geloescht+gesichert markieren
							persMgr.accessContainer(this, objId, atgId, aspId, sv, datakind, contId);
							persMgr.readContainerHeader(this);
							persMgr.restoreHeader(this, tmpContFile);
							persMgr.deleteContainerData(this);
							persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_DELETED, "1");
							persMgr.setContainerHeaderParam(this, ContainerHdr.CHP_MEDIUM_ID, mediumID);
							persMgr.writeContainerHeader(this);
						}
					}
				}
			}
			catch(PersistenceException e) {
				throw e;
			}
			finally {
				persMgr.leaveContainer(this);
				if(din != null && busyAreaEntered) din.exitBusyArea();
				tmpContFile.delete();
			}
		} // end for
	}

	/**
	 * Prueft, ob der Header des momentan im Zugriff des PersistenceManager befindlichen Containers mit dem des uebergebenen Files identisch ist. Die Reihenfolge
	 * der Header-Parameter darf unterschiedlich sein.
	 *
	 * @param file
	 *
	 * @return true, wenn Header identisch sind
	 */
	private boolean isHeaderIdentical(File file) {

		List<String> paramValues1 = new ArrayList<String>();
		List<String> paramValues2 = new ArrayList<String>();
		try {
			for(KeyValParam param : ContainerHdr.getAllParams()) {
				paramValues1.add(persMgr.getContainerHeaderParamAsString(this, param));
			}
			persMgr.leaveContainer(this);
			persMgr.accessContainerReadOnly(this, file);
			persMgr.readContainerHeader(this);
			for(KeyValParam param : ContainerHdr.getAllParams()) {
				paramValues2.add(persMgr.getContainerHeaderParamAsString(this, param));
			}
			persMgr.leaveContainer(this);
			Collections.sort(paramValues1);
			Collections.sort(paramValues2);
			return paramValues1.equals(paramValues2);
		}
		catch(Exception e) {
			logger.warning("Fehler beim Abgleich der Container-Header: " + file.getAbsolutePath(), e);
		}
		finally {
			try {
				persMgr.leaveContainer(this);
			}
			catch(PersistenceException pe) {
				// Do nothing
			}
		}
		return false;
	}
}
